import torch
from .models import getMLP
from .data import *
from .metrics import *
from torch.utils.data import DataLoader, Dataset, Subset
import json
from datetime import datetime
from time import time
import numpy as np

def get_empty_data_dict():
    ce_data = {"acc": [], "f1": [], "loss": [], "max_singular": [], "min_singular": [], "ranks": []}
    ls_data = {"acc": [], "f1": [], "loss": [], "max_singular": [], "min_singular": [], "ranks": []}
    focal_data = {"acc": [], "f1": [], "loss": [], "max_singular": [], "min_singular": [], "ranks": []}
    scor_data = {"acc": [], "f1": [], "loss": [], "max_singular": [], "min_singular": [], "ranks": []}
    return ce_data, ls_data, focal_data, scor_data

def get_empty_dist_dict():
    ce_data_dist = {"acc": [], "f1": [], "loss": [], "singulars": [], "ranks": []}
    ls_data_dist = {"acc": [], "f1": [], "loss": [], "singulars": [], "ranks": []}
    focal_data_dist = {"acc": [], "f1": [], "loss": [], "singulars": [], "ranks": []}
    scor_data_dist = {"acc": [], "f1": [], "loss": [], "singulars": [], "ranks": []}
    return ce_data_dist, ls_data_dist, focal_data_dist, scor_data_dist

def trainMLP(batch_size: int, device: str = "cuda",
          iterations: int = 1000, epochs: int = 5,
          *args, **kwargs):

    ce_data, ls_data, focal_data, scor_data = get_empty_data_dict()

    train_loader: DataLoader
    test_subset: Subset
    x_test: torch.Tensor
    y_test: torch.Tensor

    _acc = 0
    _f1 = 0
    _loss = 0
    N = 0

    try:
        begin = time()
        for i in range(iterations):
            train_loader, test_subset = getUnbalancedMNIST(batch_size, *args, **kwargs)
            x_test = test_subset.dataset.dataset.data[test_subset.indices].to(device)
            y_test = test_subset.dataset.dataset.targets[test_subset.indices].to(device)

            N = len(train_loader) * batch_size
            # instead generate test_dataset and return it

            model_ce, criterion, optimizer = getMLP(device, loss="ce")
            model_ce.train()

            for epoch in range(epochs):
                _loss = model_ce.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ce.evaluate(x_test, y_test)

            ranks, max_singulars, min_singulars = perturbations(model_ce)

            ce_data["acc"].append(_acc)
            ce_data["f1"].append(_f1)
            ce_data["loss"].append(_loss / N)
            ce_data["max_singular"].append(max_singulars)
            ce_data["min_singular"].append(min_singulars)
            ce_data["ranks"].append(ranks)

            # ---- Label Smoothing ----

            model_ls, criterion, optimizer = getMLP(device, loss="ce", label_smoothing=0.1)
            model_ls.train()
            for epoch in range(epochs):
                _loss = model_ls.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ls.evaluate(x_test, y_test)
            ranks, max_singulars, min_singulars = perturbations(model_ls)

            ls_data["acc"].append(_acc)
            ls_data["f1"].append(_f1)
            ls_data["loss"].append(_loss / N)
            ls_data["max_singular"].append(max_singulars)
            ls_data["min_singular"].append(min_singulars)
            ls_data["ranks"].append(ranks)

            # ---- Focal Model ----

            model_focal, criterion, optimizer = getMLP(device, loss="focal")
            model_focal.train()
            for epoch in range(epochs):
                _loss = model_focal.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_focal.evaluate(x_test, y_test)
            ranks, max_singulars, min_singulars = perturbations(model_focal)

            focal_data["acc"].append(_acc)
            focal_data["f1"].append(_f1)
            focal_data["loss"].append(_loss / N)
            focal_data["max_singular"].append(max_singulars)
            focal_data["min_singular"].append(min_singulars)
            focal_data["ranks"].append(ranks)

            # ---- scor model ----

            model_scor, criterion, optimizer = getMLP(device, loss="scor")
            model_scor.train()
            for epoch in range(epochs):
                _loss = model_scor.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_scor.evaluate(x_test, y_test)
            ranks, max_singulars, min_singulars = perturbations(model_scor)

            scor_data["acc"].append(_acc)
            scor_data["f1"].append(_f1)
            scor_data["loss"].append(_loss / N)
            scor_data["max_singular"].append(max_singulars)
            scor_data["min_singular"].append(min_singulars)
            scor_data["ranks"].append(ranks)

            if device == "cuda":
                torch.cuda.empty_cache()

            if i % 1 == 0:
                end = time()
                dt = end - begin
                eta = dt * iterations / (i + 1)
                print(f"Training done %{(((i + 1) * 100) / iterations):.5f} | Batch size: {batch_size}")
                print(f"ETA: {eta:.5f}s")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(ce_data, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(ls_data, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(focal_data, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(scor_data, file, ensure_ascii=False, indent=4)

    except Exception as e:
        print(f"An exception occurred: {e}")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(ce_data, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(ls_data, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(focal_data, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{timestamp}_{batch_size}.json", "w", encoding="utf-8") as file:
            json.dump(scor_data, file, ensure_ascii=False, indent=4)

def trainMLPwithAlpha(
    batch_size: int, device: str = "cuda",
    iterations: int = 1000, epochs: int = 5,
    alpha: float = 1e-04,
    *args, **kwargs
):
    _, _, _, scor_data = get_empty_data_dict()

    train_loader: DataLoader
    test_subset: Subset
    x_test: torch.Tensor
    y_test: torch.Tensor

    _acc = 0
    _f1 = 0
    _loss = 0
    N = 0

    try:
        begin = time()
        for i in range(iterations):
            train_loader, test_subset = getUnbalancedMNIST(batch_size, *args, **kwargs)
            x_test = test_subset.dataset.dataset.data[test_subset.indices].to(device)
            y_test = test_subset.dataset.dataset.targets[test_subset.indices].to(device)

            N = len(train_loader) * batch_size
            # instead generate test_dataset and return it

            model_scor, criterion, optimizer = getMLP(device=device, loss="scor", alpha=alpha)
            model_scor.train()

            for epoch in range(epochs):
                _loss = model_scor.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_scor.evaluate(x_test, y_test)

            ranks, max_singulars, min_singulars = perturbations(model_scor)

            scor_data["acc"].append(_acc)
            scor_data["f1"].append(_f1)
            scor_data["loss"].append(_loss / N)
            scor_data["max_singular"].append(max_singulars)
            scor_data["min_singular"].append(min_singulars)
            scor_data["ranks"].append(ranks)

            if device == "cuda":
                torch.cuda.empty_cache()

            if i % 1 == 0:
                end = time()
                dt = end - begin
                eta = dt * iterations / (i + 1)
                print(f"Training done %{(((i + 1) * 100) / iterations):.5f} | Batch size: {batch_size} | Alpha: {alpha}")
                print(f"ETA: {eta:.5f}s")

        timestamp = datetime.now().timestamp()
        with open(f"results/scor_data_{timestamp}_{batch_size}_{np.abs(np.log10(alpha))}.json", "w", encoding="utf-8") as file:
            json.dump(scor_data, file, ensure_ascii=False, indent=4)

    except Exception as e:
        print(f"An exception occurred: {e}")
        timestamp = datetime.now().timestamp()


def trainTargetedMLP(batch_size: int, device: str = "cuda",
          iterations: int = 1000, epochs: int = 5, reducing_factor: float = 0.1,
          *args, **kwargs):
    ce_data, ls_data, focal_data, scor_data = get_empty_data_dict()

    train_loader: DataLoader
    test_loader: DataLoader
    x_test: torch.Tensor
    y_test: torch.Tensor

    _acc = 0
    _f1 = 0
    _loss = 0
    N = 0

    reducingName = int(reducing_factor * 10)

    try:
        begin = time()
        for i in range(iterations):
            target_class = np.random.randint(0, 10)
            train_loader, test_loader = getUnbalancedMNIST9(batch_size, reducing_factor,
                                                            target_class=target_class, *args, **kwargs)
            x_test, y_test = loaderToData(test_loader)
            x_test = x_test.to(device)
            y_test = y_test.to(device)

            N = len(train_loader) * batch_size

            model_ce, criterion, optimizer = getMLP(device, loss="ce")
            model_ce.train()

            for epoch in range(epochs):
                _loss = model_ce.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ce.evaluateTargeted(x_test, y_test, target_class=target_class)

            ranks, max_singulars, min_singulars = perturbations(model_ce)

            ce_data["acc"].append(_acc)
            ce_data["f1"].append(_f1)
            ce_data["loss"].append(_loss / N)
            ce_data["max_singular"].append(max_singulars)
            ce_data["min_singular"].append(min_singulars)
            ce_data["ranks"].append(ranks)

            # ---- Label Smoothing ----

            model_ls, criterion, optimizer = getMLP(device, loss="ce", label_smoothing=0.1)
            model_ls.train()
            for epoch in range(epochs):
                _loss = model_ls.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ls.evaluateTargeted(x_test, y_test, target_class=target_class)
            ranks, max_singulars, min_singulars = perturbations(model_ls)

            ls_data["acc"].append(_acc)
            ls_data["f1"].append(_f1)
            ls_data["loss"].append(_loss / N)
            ls_data["max_singular"].append(max_singulars)
            ls_data["min_singular"].append(min_singulars)
            ls_data["ranks"].append(ranks)

            # ---- Focal Model ----

            model_focal, criterion, optimizer = getMLP(device, loss="focal")
            model_focal.train()
            for epoch in range(epochs):
                _loss = model_focal.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_focal.evaluateTargeted(x_test, y_test, target_class=target_class)
            ranks, max_singulars, min_singulars = perturbations(model_focal)

            focal_data["acc"].append(_acc)
            focal_data["f1"].append(_f1)
            focal_data["loss"].append(_loss / N)
            focal_data["max_singular"].append(max_singulars)
            focal_data["min_singular"].append(min_singulars)
            focal_data["ranks"].append(ranks)

            # ---- scor model ----

            model_scor, criterion, optimizer = getMLP(device, loss="scor")
            model_scor.train()
            for epoch in range(epochs):
                _loss = model_scor.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_scor.evaluateTargeted(x_test, y_test, target_class=target_class)
            ranks, max_singulars, min_singulars = perturbations(model_scor)

            scor_data["acc"].append(_acc)
            scor_data["f1"].append(_f1)
            scor_data["loss"].append(_loss / N)
            scor_data["max_singular"].append(max_singulars)
            scor_data["min_singular"].append(min_singulars)
            scor_data["ranks"].append(ranks)

            if device == "cuda":
                torch.cuda.empty_cache()

            if i % 1 == 0:
                end = time()
                dt = end - begin
                eta = dt * (iterations - i - 1) / (i + 1)
                print(f"Training done %{(((i + 1) * 100) / iterations):.5f} | Batch size: {batch_size}")
                print(f"ETA: {eta:.5f}s")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{iterations}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(ce_data, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{iterations}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(ls_data, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{iterations}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(focal_data, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{iterations}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(scor_data, file, ensure_ascii=False, indent=4)
    except Exception as e:
        print(f"An exception occurred: {e}")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{timestamp}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(ce_data, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{timestamp}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(ls_data, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{timestamp}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(focal_data, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{timestamp}_{batch_size}_{reducingName}.json", "w", encoding="utf-8") as file:
            json.dump(scor_data, file, ensure_ascii=False, indent=4)

def trainMLPwithdist(batch_size: int, device: str = "cuda",
          iterations: int = 1000, epochs: int = 5,
          *args, **kwargs):

    ce_data_dist, ls_data_dist, focal_data_dist, scor_data_dist = get_empty_dist_dict()

    train_loader: DataLoader
    test_subset: Subset
    x_test: torch.Tensor
    y_test: torch.Tensor

    _acc = 0
    _f1 = 0
    _loss = 0
    N = 0

    try:
        begin = time()
        for i in range(iterations):
            train_loader, test_subset = getUnbalancedMNIST(batch_size, *args, **kwargs)
            x_test = test_subset.dataset.dataset.data[test_subset.indices].to(device)
            y_test = test_subset.dataset.dataset.targets[test_subset.indices].to(device)

            N = len(train_loader) * batch_size
            # instead generate test_dataset and return it

            model_ce, criterion, optimizer = getMLP(device, loss="ce")
            model_ce.train()

            for epoch in range(epochs):
                _loss = model_ce.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ce.evaluate(x_test, y_test)

            ranks, singulars = spectral_dist(model_ce)

            ce_data_dist["acc"].append(_acc)
            ce_data_dist["f1"].append(_f1)
            ce_data_dist["loss"].append(_loss / N)
            ce_data_dist["singulars"].append(singulars)
            ce_data_dist["ranks"].append(ranks)

            # ---- Label Smoothing ----

            model_ls, criterion, optimizer = getMLP(device, loss="ce", label_smoothing=0.1)
            model_ls.train()
            for epoch in range(epochs):
                _loss = model_ls.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ls.evaluate(x_test, y_test)
            ranks, singulars = spectral_dist(model_ls)

            ls_data_dist["acc"].append(_acc)
            ls_data_dist["f1"].append(_f1)
            ls_data_dist["loss"].append(_loss / N)
            ls_data_dist["singulars"].append(singulars)
            ls_data_dist["ranks"].append(ranks)

            # ---- Focal Model ----

            model_focal, criterion, optimizer = getMLP(device, loss="focal")
            model_focal.train()
            for epoch in range(epochs):
                _loss = model_focal.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_focal.evaluate(x_test, y_test)
            ranks, singulars = spectral_dist(model_focal)

            focal_data_dist["acc"].append(_acc)
            focal_data_dist["f1"].append(_f1)
            focal_data_dist["loss"].append(_loss / N)
            focal_data_dist["singulars"].append(singulars)
            focal_data_dist["ranks"].append(ranks)

            # ---- scor model ----

            model_scor, criterion, optimizer = getMLP(device, loss="scor")
            model_scor.train()
            for epoch in range(epochs):
                _loss = model_scor.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_scor.evaluate(x_test, y_test)
            ranks, singulars = spectral_dist(model_scor)

            scor_data_dist["acc"].append(_acc)
            scor_data_dist["f1"].append(_f1)
            scor_data_dist["loss"].append(_loss / N)
            scor_data_dist["singulars"].append(singulars)
            scor_data_dist["ranks"].append(ranks)

            if device == "cuda":
                torch.cuda.empty_cache()

            if i % 1 == 0:
                end = time()
                dt = end - begin
                eta = dt * iterations / (i + 1)
                print(f"Training done %{(((i + 1) * 100) / iterations):.5f} | Batch size: {batch_size}")
                print(f"ETA: {eta:.5f}s")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ce_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ls_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(focal_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(scor_data_dist, file, ensure_ascii=False, indent=4)

    except Exception as e:
        print(f"An exception occurred: {e}")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ce_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ls_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(focal_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{timestamp}_{batch_size}_dist.json", "w", encoding="utf-8") as file:
            json.dump(scor_data_dist, file, ensure_ascii=False, indent=4)

def trainTargetedMLPwithdist(batch_size: int, device: str = "cuda",
          iterations: int = 1000, epochs: int = 5, reducing_factor: float = 0.1,
          *args, **kwargs):
    ce_data_dist, ls_data_dist, focal_data_dist, scor_data_dist = get_empty_dist_dict()

    train_loader: DataLoader
    test_loader: DataLoader
    x_test: torch.Tensor
    y_test: torch.Tensor

    _acc = 0
    _f1 = 0
    _loss = 0
    N = 0

    reducingName = int(reducing_factor * 10)

    try:
        begin = time()
        for i in range(iterations):
            target_class = np.random.randint(0, 10)
            train_loader, test_loader = getUnbalancedMNIST9(batch_size, reducing_factor,
                                                            target_class=target_class, *args, **kwargs)
            x_test, y_test = loaderToData(test_loader)
            x_test = x_test.to(device)
            y_test = y_test.to(device)

            N = len(train_loader) * batch_size

            model_ce, criterion, optimizer = getMLP(device, loss="ce")
            model_ce.train()

            for epoch in range(epochs):
                _loss = model_ce.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ce.evaluateTargeted(x_test, y_test, target_class=target_class)

            ranks, singulars = spectral_dist(model_ce)

            ce_data_dist["acc"].append(_acc)
            ce_data_dist["f1"].append(_f1)
            ce_data_dist["loss"].append(_loss / N)
            ce_data_dist["singulars"].append(singulars)
            ce_data_dist["ranks"].append(ranks)

            # ---- Label Smoothing ----

            model_ls, criterion, optimizer = getMLP(device, loss="ce", label_smoothing=0.1)
            model_ls.train()
            for epoch in range(epochs):
                _loss = model_ls.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_ls.evaluateTargeted(x_test, y_test, target_class=target_class)
            ranks, singulars = spectral_dist(model_ls)

            ls_data_dist["acc"].append(_acc)
            ls_data_dist["f1"].append(_f1)
            ls_data_dist["loss"].append(_loss / N)
            ls_data_dist["singulars"].append(singulars)
            ls_data_dist["ranks"].append(ranks)

            # ---- Focal Model ----

            model_focal, criterion, optimizer = getMLP(device, loss="focal")
            model_focal.train()
            for epoch in range(epochs):
                _loss = model_focal.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_focal.evaluateTargeted(x_test, y_test, target_class=target_class)
            ranks, singulars = spectral_dist(model_focal)

            focal_data_dist["acc"].append(_acc)
            focal_data_dist["f1"].append(_f1)
            focal_data_dist["loss"].append(_loss / N)
            focal_data_dist["singulars"].append(singulars)
            focal_data_dist["ranks"].append(ranks)

            # ---- scor model ----

            model_scor, criterion, optimizer = getMLP(device, loss="scor")
            model_scor.train()
            for epoch in range(epochs):
                _loss = model_scor.model_training(train_loader, criterion, optimizer, device)

            _acc, _f1 = model_scor.evaluateTargeted(x_test, y_test, target_class=target_class)
            ranks, singulars = spectral_dist(model_scor)

            scor_data_dist["acc"].append(_acc)
            scor_data_dist["f1"].append(_f1)
            scor_data_dist["loss"].append(_loss / N)
            scor_data_dist["singulars"].append(singulars)
            scor_data_dist["ranks"].append(ranks)

            if device == "cuda":
                torch.cuda.empty_cache()

            if i % 1 == 0:
                end = time()
                dt = end - begin
                eta = dt * (iterations - i - 1) / (i + 1)
                print(f"Training done %{(((i + 1) * 100) / iterations):.5f} | Batch size: {batch_size}")
                print(f"ETA: {eta:.5f}s")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{iterations}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ce_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{iterations}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ls_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{iterations}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(focal_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{iterations}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(scor_data_dist, file, ensure_ascii=False, indent=4)
    except Exception as e:
        print(f"An exception occurred: {e}")

        timestamp = datetime.now().timestamp()

        with open(f"results/ce_data_{timestamp}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ce_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/ls_data_{timestamp}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(ls_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/focal_data_{timestamp}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(focal_data_dist, file, ensure_ascii=False, indent=4)

        with open(f"results/scor_data_{timestamp}_{batch_size}_{reducingName}_dist.json", "w", encoding="utf-8") as file:
            json.dump(scor_data_dist, file, ensure_ascii=False, indent=4)





import torch.nn as nn
import torch
from torch.utils.data import DataLoader
import torch.optim as optim
from torchvision.models.convnext import (
    convnext_tiny,
    convnext_small,
    convnext_base,
    convnext_large,
    ConvNeXt_Tiny_Weights,
    ConvNeXt_Small_Weights,
    ConvNeXt_Base_Weights,
    ConvNeXt_Large_Weights,
)

from .losses import *

from tqdm import tqdm
from sklearn.metrics import f1_score

class MLP(nn.Module):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sequence = nn.Sequential(
            nn.Linear(784, 128),
            nn.ReLU(),
            nn.Linear(128, 10)
        )

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        x = x.view(-1, 784)
        return self.sequence(x)

    def predict(self, x):
        x = x.view(-1, 784)
        return torch.argmax(self.softmax(self.sequence(x)), dim=-1)

    def model_training(self, train_loader: DataLoader,
                       criterion: nn.Module, optimizer: optim.Optimizer,
                       device: str = "cuda"):

        _loss = 0

        N: int = next(iter(train_loader))[0].shape[0]
        pbar = tqdm(train_loader, total=len(train_loader), desc=f"Training")

        for i, (x, y) in enumerate(pbar):
            optimizer.zero_grad()
            x = x.to(device)
            y = y.to(device)
            out = self(x)
            loss = criterion(out, y)
            _loss += loss.item()
            loss.backward()
            optimizer.step()
            pbar.set_postfix({"Loss": _loss / ((i + 1) * N)})

        if device == "cuda":
            torch.cuda.empty_cache()
        return _loss


    @torch.no_grad()
    def evaluate(self, test_data: torch.Tensor, test_labels: torch.Tensor,
                 n_parts: int = 5, empty_cache: bool = True):

        N = test_data.shape[0] // n_parts
        temp_out: torch.Tensor
        acc = 0
        f1_scores = 0

        for i in range(n_parts):
            temp_out = self.predict(test_data[N * i:N * (i + 1)].type(torch.float32))
            acc += (torch.count_nonzero(temp_out == test_labels[N * i:N * (i + 1)]) / N).item()

            f1_scores += f1_score(test_labels[N * i:N * (i + 1)].type(torch.uint8).cpu().numpy().ravel(),
                                  temp_out.type(torch.uint8).cpu().numpy().ravel(),
                                  average="macro", zero_division=0)

            if empty_cache:
                torch.cuda.empty_cache()

        return acc / n_parts, f1_scores / n_parts

    @torch.no_grad()
    def evaluateTargeted(self, test_data: torch.Tensor, test_labels: torch.Tensor,
                         n_parts: int = 5, empty_cache: bool = True, target_class: int = 0):

        N = test_data.shape[0] // n_parts
        temp_out: torch.Tensor
        acc = 0
        f1_scores = 0

        for i in range(n_parts):
            mask = (test_labels[N * i:N * (i + 1)].type(torch.uint8) != target_class)
            temp_out = self.predict(test_data[N * i:N * (i + 1)][mask].type(torch.float32))
            tests = test_labels[N * i:N * (i + 1)][mask]

            acc += (torch.count_nonzero(temp_out == tests) / N).item()

            f1_scores += f1_score(tests.type(torch.uint8).cpu().numpy().ravel(),
                                  temp_out.type(torch.uint8).cpu().numpy().ravel(),
                                  average="macro", zero_division=0)

            if empty_cache:
                torch.cuda.empty_cache()

        return acc / n_parts, f1_scores / n_parts

class ResidualBlock(nn.Module):

    def __init__(self, in_channels: int, out_channels: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.sequence = nn.Sequential(
            nn.Conv2d(self.in_channels, self.out_channels, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU(),
            nn.Conv2d(self.out_channels, self.out_channels, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU(),
            nn.Conv2d(self.out_channels, self.out_channels, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU(),
            nn.Conv2d(self.out_channels, self.out_channels, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(self.out_channels),
        )

    def forward(self, x):
        return self.sequence(x) + x

class ResNet(nn.Module):

    """
        This is the basic ResNet structure used for CIFAR-10 tests.
        This excludes CIFAR-10-LT.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.sequence = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            ResidualBlock(32, 32),
            nn.ReLU(),
            ResidualBlock(32, 32),
            nn.ReLU(),
            ResidualBlock(32, 32),
            nn.ReLU(),
            ResidualBlock(32, 32),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Flatten(),
            nn.Linear(64 * 32 * 32, 64 * 32),
            nn.Dropout(0.5),
            nn.ReLU(),
            nn.Linear(64 * 32, 10)
        )

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        out = self.sequence(x)
        return out

    def predict(self, x):
        x = x.reshape(-1, 3, 32, 32)
        out = self.softmax(self.sequence(x))
        return torch.argmax(out, dim=-1)

    def model_training(self, train_loader: DataLoader,
                       criterion: nn.Module, optimizer: optim.Optimizer,
                       device: str = "cuda", N=64):

        _loss = 0
        pbar = tqdm(train_loader, total=len(train_loader), desc=f"Training")

        for i, (x, y) in enumerate(pbar):
            optimizer.zero_grad()
            x = x.to(device)
            y = y.to(device)
            out = self(x)
            #print(out.shape, y.shape)
            loss = criterion(out, y)
            loss.backward()
            _loss += loss.item()
            optimizer.step()

            pbar.set_postfix({"Loss": _loss / ((i + 1) * N)})

        if device == "cuda":
            torch.cuda.empty_cache()
        return _loss

    @torch.no_grad()
    def evaluate(self, test_data: torch.Tensor, test_labels: torch.Tensor,
                 n_parts: int = 1, empty_cache: bool = True):

        N = test_data.shape[0] // n_parts
        temp_out: torch.Tensor
        acc = 0
        f1_scores = 0

        for i in range(n_parts):
            temp_out = self.predict(test_data[N * i:N * (i + 1)].type(torch.float32))
            acc += (torch.count_nonzero(temp_out == test_labels[N * i:N * (i + 1)]) / N).item()

            f1_scores += f1_score(test_labels[N * i:N * (i + 1)].type(torch.uint8).cpu().numpy().ravel(),
                                  temp_out.type(torch.uint8).cpu().numpy().ravel(),
                                  average="macro", zero_division=0)

            if empty_cache:
                torch.cuda.empty_cache()

        return acc / n_parts, f1_scores / n_parts

    @torch.no_grad()
    def evaluateTargeted(self, test_data: torch.Tensor, test_labels: torch.Tensor,
                 n_parts: int = 1, empty_cache: bool = True, target_class: int = 0):

        N = test_data.shape[0] // n_parts
        temp_out: torch.Tensor
        acc = 0
        f1_scores = 0

        for i in range(n_parts):
            mask = (test_labels[N * i:N * (i + 1)].type(torch.uint8) != target_class)
            temp_out = self.predict(test_data[N * i:N * (i + 1)][mask].type(torch.float32))
            tests = test_labels[N * i:N * (i + 1)][mask]

            acc += (torch.count_nonzero(temp_out == tests) / N).item()

            f1_scores += f1_score(tests.type(torch.uint8).cpu().numpy().ravel(),
                                  temp_out.type(torch.uint8).cpu().numpy().ravel(),
                                  average="macro", zero_division=0)

            if empty_cache:
                torch.cuda.empty_cache()

        return acc / n_parts, f1_scores / n_parts

class ConvNeXtBase(nn.Module):

    def __init__(self, size: str = "tiny", *args, **kwargs):
        """
            This is the base class for ConvNeXt applications,
            where ConvNeXt is needed as the feature extractor.

            This class, under its "feature_extractor" attribute,
            has the vector outputs of selected model size.

            Args:
                size (str): The size of the ConvNeXt to be used.
                    Must be one of ['tiny', 'small', 'base', 'large'].
        """

        super().__init__(*args, **kwargs)
        self.size = size.lower()
        self.channels = None

        if self.size == "tiny":
            weights = ConvNeXt_Tiny_Weights.DEFAULT
            backbone = convnext_tiny(weights=weights)
        elif self.size == "small":
            weights = ConvNeXt_Small_Weights.DEFAULT
            backbone = convnext_small(weights=weights)
        elif self.size == "base":
            weights = ConvNeXt_Base_Weights.DEFAULT
            backbone = convnext_base(weights=weights)
        elif self.size == "large":
            weights = ConvNeXt_Large_Weights.DEFAULT
            backbone = convnext_large(weights=weights)
        else:
            raise ValueError(f"size argument must be one of ['tiny', 'small', 'base', 'large']")

        self.feature_extractor = nn.Sequential(*list(backbone.children())[:-1])

    def forward(self, x):
        return self.feature_extractor(x)

class ConvNeXtClassification(ConvNeXtBase):

    def __init__(self, size: str = "tiny", num_classes: int = 1000, *args, **kwargs):
        """
            The classification oriented extension of ConvNeXtBase class. This class,
            adds a single hidden layer classification head with "num_classes" outputs.

            Args:
                size (str): The size of the ConvNeXt base to be used.
                    Must be one of ['tiny', 'small', 'base', 'large'].
                    Defaults to 'tiny'.

                num_classes (int): Number of classes to predict. Defaults
                    to 1000 for ImageNet.
        """

        super().__init__(size=size, *args, **kwargs)

        self.input_dims = 768 if self.size == "tiny" or self.size == "small" else \
            1024 if self.size == "base" else 1536

        self.num_classes = num_classes

        self.head = nn.Sequential(
            nn.Linear(self.input_dims, 512),
            nn.ReLU(),
            nn.Linear(512, self.num_classes)
        )

    def forward(self, x, opl: bool = False):
        features = super().forward(x)
        features = features.view(features.size(0), -1)
        if opl:
            return self.head(features), features
        return self.head(features)  # This will still be returned as a tuple? Because Python?

class FaceVerificationModel(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()

        weights = ConvNeXt_Tiny_Weights.DEFAULT
        backbone = convnext_tiny(weights=weights)

        self.feature_extractor = nn.Sequential(*list(backbone.children())[:-1])

        self.fc = nn.Sequential(
            nn.Linear(768, 512),
            nn.ReLU(),
            nn.Linear(512, num_classes)
        )

    def forward(self, img1):
        f1 = self.feature_extractor(img1)

        f1 = f1.view(f1.shape[0], -1)  # Flatten
        out = self.fc(f1)  # Pass through classification head
        return out, f1

def getMLP(device: str = "cuda", loss: str = "ce", *args, **kwargs):
    model = MLP().to(device)
    if loss == "ce":  # Label smoothing is included here
        _loss = nn.CrossEntropyLoss(*args, **kwargs)
    elif loss == "scor":
        _loss = SCoR(*args, **kwargs, device=device)
    elif loss == "focal":
        _loss = FocalLoss(*args, **kwargs)
    else:
        raise

    return model, _loss, optim.Adam(model.parameters())

def getResNet(device: str, loss: str = "ce", *args, **kwargs):

    model = ResNet().to(device)

    if loss == "ce":  # Label smoothing is included here
        _loss = nn.CrossEntropyLoss(*args, **kwargs)
    elif loss == "scor":
        _loss = SCoR(*args, **kwargs, device=device)
    elif loss == "focal":
        _loss = FocalLoss(*args, **kwargs)
    else:
        raise

    return model, _loss, optim.Adam(model.parameters())



from torchvision import transforms
from torchvision import datasets
from torchvision.datasets import MNIST, CIFAR10, Flowers102, OxfordIIITPet, DTD, STL10, ImageFolder
from torch.utils.data import DataLoader, random_split, Subset
import torch

import numpy as np
import os
import shutil

transform_ = transforms.Compose([
    transforms.Resize((224, 224)),
    transforms.RandomHorizontalFlip(),
    transforms.RandomRotation(15),
    transforms.ColorJitter(brightness=0.2, contrast=0.2, saturation=0.2),
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
])

normalTransform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.,), (1.,))
])

uniformTransform = transforms.Compose([
    transforms.RandomHorizontalFlip(),
    transforms.RandomCrop(32, padding=4),
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2470, 0.2435, 0.2616]),
])

def get_cifar10(transform=None,
                batch_size: int = 64,
                shuffle: bool = True,
                num_workers: int = 0,
                drop_last: bool = False):

    train_set = CIFAR10(root='./data', train=True, download=True, transform=transform)
    test_set = CIFAR10(root='./data', train=False, download=True, transform=transform)

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)

    return train_loader, test_loader

def get_mnist(transform=None,
              batch_size: int = 64,
              shuffle: bool = True,
              num_workers: int = 0,
              drop_last: bool = False):

    train_set = MNIST(root='./data', train=True, download=True, transform=transform)
    test_set = MNIST(root='./data', train=False, download=True, transform=transform)

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)

    return train_loader, test_loader

def get_flowers102(transform=transform_,
                   batch_size: int = 32,
                   shuffle: bool = True,
                   num_workers: int = 0,
                   drop_last: bool = False):

    train_set = Flowers102(root="./data", split="train", download=True, transform=transform)
    test_set = Flowers102(root="./data", split="test", download=True, transform=transform)

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)

    return train_loader, test_loader

def get_oxfordpets(transform=transform_,
                   batch_size: int = 32,
                   shuffle: bool = True,
                   num_workers: int = 0,
                   drop_last: bool = False):
    train_set = OxfordIIITPet(root="./data", split="trainval", download=True, transform=transform)
    test_set = OxfordIIITPet(root="./data", split="test", download=True, transform=transform)

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                              drop_last=drop_last)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                             drop_last=drop_last)

    return train_loader, test_loader

def get_stl10(transform=transform_,
              batch_size: int = 128,
              shuffle: bool = True,
              num_workers: int = 0,
              drop_last: bool = False):

    train_set = STL10(root="./data", split="train", download=True, transform=transform)
    test_set = STL10(root="./data", split="test", download=True, transform=transform)

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                              drop_last=drop_last)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                             drop_last=drop_last)

    return train_loader, test_loader

def get_dtd(transform=transform_,
            batch_size: int = 32,
            shuffle: bool = True,
            num_workers: int = 0,
            drop_last: bool = False):

    train_set = DTD(root="./data", split="train", download=True, transform=transform)
    test_set = DTD(root="./data", split="test", download=True, transform=transform)

    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                              drop_last=drop_last)
    test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                             drop_last=drop_last)

    return train_loader, test_loader

def get_lfw(lfw_root: str,
            transform=transform_,
            batch_size: int = 128,
            trainset_size: float = 0.8,
            shuffle: bool = True,
            num_workers: int = 0,
            drop_last: bool = False,
            exclude: int = 5):

    for folder in os.listdir(lfw_root):
        folder_path = os.path.join(lfw_root, folder)
        if os.path.isdir(folder_path):  # Check if it's a directory
            num_images = len(os.listdir(folder_path))  # Count images
            if num_images < exclude:
                shutil.rmtree(folder_path)  # Delete the folder
                print(f"Deleted: {folder_path} (Had {num_images} images)")

    dataset = ImageFolder(lfw_root, transform=transform)
    train_size = int(trainset_size * len(dataset))
    test_size = len(dataset) - train_size

    train_dataset, test_dataset = random_split(dataset, [train_size, test_size])
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)
    return train_loader, test_loader

def loadMNIST(transform: transforms.Compose = normalTransform, batch_size: int = 64):
    train_dataset = datasets.MNIST(root="data", train=True, transform=transform, download=True)
    test_dataset = datasets.MNIST(root="data", train=False, transform=transform, download=True)

    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(dataset=test_dataset, shuffle=True)

    return train_loader, test_loader

def getUnbalancedMNIST(batch_size: int, split: float = 0.8, transform: transforms.Compose = normalTransform,
                       reducing_factor: float = 0.05, target_class: int = None):

    if target_class is None:
        target_class = np.random.randint(0, 10)

    train_dataset = datasets.MNIST(root="data", train=True, transform=transform, download=True)
    class_indices = {i: (train_dataset.targets == i).nonzero(as_tuple=True)[0] for i in range(10)}

    # randomize the order of data for each class
    for key, value in class_indices.items():
        perm = torch.randperm(len(value))
        class_indices[key] = value[perm]

    num_samples = len(class_indices[target_class])
    reduced_indices = class_indices[target_class][:int(reducing_factor * num_samples)]

    unbalanced_indices = reduced_indices.tolist()
    for i in range(10):
        if target_class == i:
            continue
        unbalanced_indices.extend(class_indices[i].tolist())

    unbalanced_dataset = Subset(train_dataset, unbalanced_indices)

    train_size = int(split * len(unbalanced_dataset))
    test_size = len(unbalanced_dataset) - train_size

    train_dataset, test_dataset = random_split(unbalanced_dataset, [train_size, test_size])

    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)

    return train_loader, test_dataset


def getUnbalancedMNIST9(
        batch_size: int,
        split: float = 0.8,
        transform: transforms.Compose = normalTransform,
        reducing_factor: float = 0.1,
        target_class: int = None
):

    if target_class is None:
        target_class = np.random.randint(0, 10)

    train_dataset = datasets.MNIST(root="data", train=True, transform=transform, download=True)
    class_indices = {i: (train_dataset.targets == i).nonzero(as_tuple=True)[0] for i in range(10)}

    for key in class_indices:
        class_indices[key] = class_indices[key][torch.randperm(len(class_indices[key]))]

    unbalanced_indices = []
    for i in range(10):
        if target_class == i:
            unbalanced_indices.extend(class_indices[i].tolist())
        else:
            num_samples = len(class_indices[i])
            unbalanced_indices.extend(class_indices[i][:int(reducing_factor * num_samples)].tolist())

    unbalanced_dataset = Subset(train_dataset, unbalanced_indices)

    train_size = int(split * len(unbalanced_dataset))
    test_size = len(unbalanced_dataset) - train_size
    train_dataset, test_dataset = random_split(unbalanced_dataset, [train_size, test_size])

    # Create DataLoader instances
    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)

    return train_loader, test_loader

def loadCIFAR10(transform: transforms.Compose = uniformTransform, batch_size: int = 64):

    train_dataset = datasets.CIFAR10(
        root='./data',
        train=True,
        download=True,
        transform=transform
    )

    test_dataset = datasets.CIFAR10(
        root='./data',
        train=False,
        download=True,
        transform=transform
    )

    train_loader = DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        shuffle=True
    )

    test_loader = DataLoader(
        dataset=test_dataset,
        shuffle=True
    )

    return train_loader, test_loader


def getUnbalancedCIFAR(
        batch_size: int,
        split: float = 0.8,
        transform: transforms.Compose = uniformTransform,
        reducing_factor: float = 0.05,
        target_class: int = None
):

    if target_class is None:
        target_class = np.random.randint(0, 10)


    train_dataset = datasets.CIFAR10(root="data", train=True, transform=transform, download=True)

    class_indices = {i: (torch.tensor(train_dataset.targets) == i).nonzero(as_tuple=True)[0] for i in range(10)}


    for key in class_indices:
        class_indices[key] = class_indices[key][torch.randperm(len(class_indices[key]))]


    num_samples = len(class_indices[target_class])
    reduced_indices = class_indices[target_class][:int(reducing_factor * num_samples)]


    unbalanced_indices = reduced_indices.tolist()
    for i in range(10):
        if target_class == i:
            continue
        unbalanced_indices.extend(class_indices[i].tolist())


    unbalanced_dataset = Subset(train_dataset, unbalanced_indices)


    train_size = int(split * len(unbalanced_dataset))
    test_size = len(unbalanced_dataset) - train_size
    train_dataset, test_dataset = random_split(unbalanced_dataset, [train_size, test_size])

    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_dataset, shuffle=False)
    return train_loader, test_loader

def getUnbalancedCIFAR9(
        batch_size: int,
        split: float = 0.8,
        transform: transforms.Compose = uniformTransform,
        reducing_factor: float = 0.05,
        target_class: int = None
):

    if target_class is None:
        target_class = np.random.randint(0, 10)


    train_dataset = datasets.CIFAR10(root="./data", train=True, transform=transform, download=True)

    class_indices = {i: (torch.tensor(train_dataset.targets) == i).nonzero(as_tuple=True)[0] for i in range(10)}

    for key in class_indices:
        class_indices[key] = class_indices[key][torch.randperm(len(class_indices[key]))]

    unbalanced_indices = []
    for i in range(10):
        if target_class == i:
            unbalanced_indices.extend(class_indices[i].tolist())
        else:
            num_samples = len(class_indices[i])
            unbalanced_indices.extend(class_indices[i][:int(reducing_factor * num_samples)].tolist())

    unbalanced_dataset = Subset(train_dataset, unbalanced_indices)

    train_size = int(split * len(unbalanced_dataset))
    test_size = len(unbalanced_dataset) - train_size
    train_dataset, test_dataset = random_split(unbalanced_dataset, [train_size, test_size])

    # Create DataLoader instances
    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)

    return train_loader, test_loader

def loaderToData(dataloader: DataLoader):
    x_list = []
    y_list = []

    for x_batch, y_batch in dataloader:
        x_list.append(x_batch)
        y_list.append(y_batch)

    return torch.cat(x_list, dim=0), torch.cat(y_list, dim=0)



import torch
import torch.nn as nn
import torch.nn.functional as F

class SCoR(nn.Module):

    def __init__(self, alpha: float = 1e-04, *args, **kwargs):
        """
            Stochastic Covariance Regularisation. Default alpha is 1e-04.

        """
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.crossEntropy = nn.CrossEntropyLoss()

    def forward(self, y_pred, y_true):
        _loss = self.crossEntropy(y_pred, y_true)

        if y_pred.shape[0] % 2 == 0:
            perm = torch.randperm(y_pred.shape[0]).to(y_pred.device)
            randomized = y_pred.index_select(0, perm)
            first_half = randomized[y_pred.shape[0] // 2:, :]  # Indeed this is the second half
            second_half = randomized[:y_pred.shape[0] // 2, :]  # and this is the first half
            _loss_svr = self.alpha * torch.norm(first_half.T @ second_half, p="fro")
        else:
            _loss_svr = self.alpha * torch.norm(y_pred.T @ y_pred, p="fro")

        return _loss + _loss_svr

class SCoRandLS(nn.Module):

    def __init__(self, alpha: float = 1e-04, *args, **kwargs):
        """
            Combination of SCoR with Label Smoothing (LS). LS uses p=0.1, SCoR uses
            the default alpha value of 1e-04.
        """
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.crossEntropy = nn.CrossEntropyLoss(label_smoothing=0.1)

    def forward(self, y_pred, y_true):
        _loss = self.crossEntropy(y_pred, y_true)

        if y_pred.shape[0] % 2 == 0:
            perm = torch.randperm(y_pred.shape[0]).to(y_pred.device)
            randomized = y_pred.index_select(0, perm)
            first_half = randomized[y_pred.shape[0] // 2:, :]  # Indeed this is the second half
            second_half = randomized[:y_pred.shape[0] // 2, :]  # and this is the first half
            _loss_svr = self.alpha * torch.norm(first_half.T @ second_half, p="fro")
        else:
            _loss_svr = self.alpha * torch.norm(y_pred.T @ y_pred, p="fro")

        return _loss + _loss_svr

class OPL(nn.Module):
    def __init__(self, epsilon=1e-6, alpha: float = 1, gamma: float = 0.5):
        """
            Orthogonal Projection Los, directly implemented from the Torch
            pseudocode of the paper.

            The default configuration is the best hyperparameter configuration
            reported within the paper.
        """
        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.cross_entropy = nn.CrossEntropyLoss()
        self.epsilon = epsilon

    def forward(self, logits, labels, features):
        batch_size, feature_dim = features.shape
        ce_loss_ = self.cross_entropy(logits, labels)

        features = F.normalize(features, p=2, dim=1)

        mask = torch.eq(labels, labels.t())
        eye = torch.eye(mask.shape[0], device=features.device)
        mask_pos = mask.masked_fill(eye.type(torch.bool), 0)
        mask_neg = ~mask

        dot_prod = torch.matmul(features, features.t())
        pos_total = (mask_pos * dot_prod).sum()
        neg_total = torch.abs(mask_neg * dot_prod).sum()
        pos_mean = pos_total / (mask_pos.sum() + self.epsilon)
        neg_mean = neg_total / (mask_neg.sum() + self.epsilon)

        loss = (1.0 - pos_mean) + self.gamma * neg_mean

        return ce_loss_ + self.alpha * loss / batch_size

class FocalLoss(nn.Module):

    def __init__(self, gamma: float = 2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gamma = gamma

    def forward(self, y_pred, y_true):
        probs = F.softmax(y_pred, dim=-1)
        targets_one_hot = F.one_hot(y_true, num_classes=y_pred.shape[-1]).float()
        pt = torch.sum(targets_one_hot * probs, dim=-1)
        focal_weight = (1 - pt) ** self.gamma
        log_pt = F.log_softmax(y_pred, dim=-1)
        log_pt = torch.sum(targets_one_hot * log_pt, dim=-1)
        return -torch.sum(focal_weight * log_pt)  # Sum over batches
        #return - torch.sum(((y_true - y_pred) ** self.gamma) * torch.log(y_pred))

# Use nn.CrossEntropyLoss(label_smoothing=0.1) for label smoothing



import torch
import torch.nn as nn
from itertools import chain

def perturbations(model: nn.Module, epsilon: float = 1e-07):
    ranks = []
    max_singulars = []
    min_singulars = []
    s: torch.Tensor

    # Model must have all layers in self.sequence
    for name, param in model.sequence.named_parameters():
        if "bias" in name:
            continue

        s = torch.linalg.svdvals(param)
        ranks.append(torch.sum(s > epsilon).item())
        max_singulars.append(s[0].item())
        min_singulars.append(s[-1].item())

    return ranks, max_singulars, min_singulars

def spectral_dist(model: nn.Module, epsilon: float = 1e-07):
    ranks = []
    singulars = []
    s: torch.Tensor

    for name, param in model.sequence.named_parameters():
        if "bias" in name:
            continue

        s = torch.linalg.svdvals(param)
        ranks.append(torch.sum(s > epsilon).item())
        singulars.append(s.tolist())

    return ranks, list(chain(*singulars))

def perturbationsMultiple(model: nn.Module, epsilon: float = 1e-07):
    ranks = []
    max_singulars = []
    min_singulars = []
    s: torch.Tensor

    # Model must have all layers in self.sequence
    for name, param in model.sequence.named_parameters():
        if "bias" in name or param.ndim < 2:
            continue

        s = torch.linalg.svdvals(param)
        ranks.append(torch.sum(s > epsilon).item())
        max_singulars.append(torch.mean(s[0]).item())
        min_singulars.append(torch.mean(s[-1]).item())

    return ranks, max_singulars, min_singulars
